library(magrittr)
library(tidyverse)
library(tidygraph)
library(JuliaCall)

julia_setup()
julia_source("multipop_MANTIS.jl")

## structural parameters
struct <- c(3,3)
strains <- expand.grid(lapply(struct, seq, from=1, by=1) %>%
                         setNames(str_c("locus", 1:length(.))))
n_populations <- 1
file_base <- str_c("MANTIS_bifurcation_diagram_", str_c(struct, collapse=""))

## dynamical parameters
# beta <- matrix(c(40), nrow=n_populations, ncol=prod(struct))
beta <- 40
gamma <- seq(0.01, 0.99, length.out=1000)
sigma <- 10
# chi <- diag(-0.02, n_populations)
chi <- matrix(c(-0.02, 0, 0.02, -0.02), n_populations, n_populations)

nn <- expand.grid(str_c("Population_", 1:n_populations),
                  str_c("Strain_", apply(strains, 1, str_c, collapse="")),
                  c("y", "z", "w")) %>% apply(1, str_c, collapse=".")

## run the simulation
tmp <- lapply(gamma, function(gg) {
  lapply(0, function(rseed){
    set.seed(rseed)
    initial_conditions <- runif(prod(struct) * n_populations) %>%
      matrix(n_populations, prod(struct)) %>%
      apply(1, . %>% {. / (5 * sum(.))}) %>%
      t()

    # mantis_output <- run_multipop_mantis(struct, initial_conditions, t_max=2500, t_step=0.01,
    #                                      beta=beta, gamma=gg, sigma=sigma, chi=chi)
    # timeseries <- mantis_output$timeseries %>%
    #   as.data.frame() %>%
    #   as_tibble() %>%
    #   gather("details", "prevalence", -time) %>%
    #   separate(details, c("population", "strain", "equation"), "\\.") %>%
    #   mutate(gamma=gg, rep=rseed)

    mantis_output_jl <- julia_call("runMANTIS", strainstructure=struct, tstep=0.1, tmax=15000,
                                   beta=beta, gamma=gg, sigma=sigma, chi=chi, initvals=initial_conditions)
    if (length(mantis_output_jl$timeseries) == prod(struct) * n_populations * 3 + 1) {
      return(mantis_output_jl$timeseries %>% set_colnames(c(nn, "time")) %>% filter(time > 0.9 * max(time)) %>% mutate(gamma=gg, rep=rseed))
    } else return(NULL)
  }) %>% bind_rows()
}) %>% bind_rows()

plotdata <- lapply(tmp %>% select(-time, -gamma, -rep) %>% names() %>% str_subset("\\.z$"), function(zzz) {
  tmp %>%
    select(time, gamma, rep, zzz) %>%
    gather("details", "prevalence", -time, -gamma, -rep) %>%
    separate(details, c("population", "strain", "equation"), "\\.") %>%
    mutate(prevalence = round(prevalence, 6),
           variable = factor(equation, levels=c("z", "w", "y"),
                             labels=c("specific immunity",
                                      "cross-reactive immunity",
                                      "currently infectious"))) %>%
    filter(time > 0.9 * max(time),
           equation == "z") %>%
    select(-time, -equation) %>%
    group_by(population, strain, gamma, rep) %>% distinct() %>%
    filter((prevalence > lead(prevalence, default=10) & prevalence > lag(prevalence, default=10)) |
             (prevalence < lead(prevalence, default=10) & prevalence < lag(prevalence, default=10))) %>%
    ungroup() %>%
    distinct() %>%
    count(population, strain, gamma, prevalence)
}) %>% bind_rows()

p <- ggplot(plotdata) +
  aes(x=gamma, y=prevalence, colour=as.factor(n)) +
  geom_point() +
  scale_colour_viridis_d() +
  scale_x_continuous(expand=expand_scale()) +
  facet_grid(strain~population, scales="free") +
  theme_bw() + theme(legend.position="none")
p
# ggsave(p, filename=str_c("../Figures/", file_base, "_.pdf"), width=4, height=prod(struct))
