library(deSolve)
library(glue)
library(assertthat)
library(magrittr)
library(tidyverse)
library(JuliaCall)
library(MANTIS)
library(facetscales)

julia_setup()
julia_source("multipop_MANTIS.jl")
source("multipop_MANTIS_in_R.R")

my_cols <- c("#016394", "#f74700", "#b6003b", "#005342")

scales_y <- list(
  `currently infectious`    = scale_y_continuous(limits = c(0, NA)),
  `specific immunity`       = scale_y_continuous(limits = c(0, 1)),
  `cross-reactive immunity` = scale_y_continuous(limits = c(0, 1))
)

count_unique_local_minima <- function(vec, digits=NA) {
  if (!is.na(digits)) vec <- round(vec, digits)
  # pick out the local minima
  minima <- vec[vec < lead(vec) & vec < lag(vec)]
  # count how many unique minima (greater than 0) there are
  number_unique_minima <- sum(unique(minima) > 0, na.rm=TRUE)
  # if no minima -- the timeseries is a constant, so check whether it is 0 (i.e. extinct) or not
  # otherwise return the number of minima
  if (number_unique_minima == 0 & all(vec == 0)) return(-1) else return(number_unique_minima)
}

## structural parameters
set.seed(0)
struct <- c(2,2)
strains <- expand.grid(lapply(struct, seq, from=1, by=1) %>%
                         setNames(str_c("locus", 1:length(.))))
n_populations <- 1

## dynamical parameters
beta <- 2000                # Infection rate
gamma <- 0.5             # partial cross-protective immunity (cpi)
sigma <- 500               # recovery rate
mu <- 0.02                   # disease induced mortality
# delta <- NA               # increase in cpi per allele (not yet implemented)
# epsilon <- 0              # seasonality
chi <- diag(0, n_populations)
# chi <- matrix(c(-0.05, 0, 0.05, 0), 2, 2)

initial_conditions <- runif(prod(struct) * n_populations) %>%
  matrix(n_populations, prod(struct)) %>%
  apply(1, . %>% {. / (5 * sum(.))}) %>%
  t()

## run the simulation
run_julia <- function() {
  # set.seed(0)
  # initial_conditions <- runif(prod(struct) * n_populations) %>%
  #   matrix(n_populations, prod(struct)) %>%
  #   apply(1, . %>% {. / (5 * sum(.))}) %>%
  #   t()
  julia_call("runMANTIS", strainstructure=struct, tstep=seq(0, 1000, 0.1), tmax=1000,
             beta=beta, gamma=gamma, sigma=sigma, mu=mu, chi=chi, initvals=initial_conditions)
}

# run_R <- function() run_multipop_mantis(struct, initial_conditions, t_max=1000, t_step=0.1,
#                                        beta=beta, gamma=gamma, mu=mu, sigma=sigma, chi=chi)

run_C <- function() {
  # set.seed(0)
  # initial_conditions <- runif(prod(struct) * n_populations) %>%
  #   matrix(n_populations, prod(struct)) %>%
  #   apply(1, . %>% {. / (5 * sum(.))}) %>%
  #   t()
  runMANTIS(epiStruc=struct, initCondY=initial_conditions[1,],
            tMax=1000, tObsPer=10, tStep=0.01,
            beta=beta, gamma=gamma, sigma=sigma, mu=0.02,
            initCondZ=initial_conditions[1,], initCondW=pmin(1, 3*initial_conditions[1,]))
}

# microbenchmark::microbenchmark(run_julia(), run_C())

mantis_output_jl <- run_julia()
# mantis_output_R <- run_R()
mantis_output_C <- run_C()

## clean the output
timeseries_jl <- mantis_output_jl$timeseries %>%
  set_colnames(c(expand.grid(str_c("Population_", 1:n_populations),
                             str_c("Strain_", apply(strains, 1, str_c, collapse="")),
                             c("y", "z", "w")) %>% apply(1, str_c, collapse="."),
                 "time")) %>%
  gather("details", "prevalence", -time) %>%
  separate(details, c("population", "strain", "equation"), "\\.") %>%
  mutate(variable = factor(equation, levels=c("z", "w", "y"),
                           labels=c("specific immunity",
                                    "cross-reactive immunity",
                                    "currently infectious")))

# timeseries_R <- mantis_output_R$timeseries %>%
#   as.data.frame() %>% # necessary to remove attributes from ode
#   as_tibble() %>%
#   gather("details", "prevalence", -time) %>%
#   separate(details, c("population", "strain", "equation"), "\\.") %>%
#   mutate(variable = factor(equation, levels=c("z", "w", "y"),
#                            labels=c("specific immunity",
#                                     "cross-reactive immunity",
#                                     "currently infectious")))

timeseries_C <- as_tibble(mantis_output_C$Y %>%
                            set_colnames(expand.grid(str_c("Population_", 1:n_populations),
                                                     str_c("Strain_", lapply(struct, seq, from=1, by=1) %>%
                                                             expand.grid() %>%
                                                             as.data.frame() %>%
                                                             unite("ss", sep="") %>%
                                                             use_series(ss)),
                                                     c("z", "w", "y")) %>%
                                           unite("x", sep=".") %>% use_series(x))) %>%
  mutate(time=mantis_output_C$T) %>%
  gather("details", "prevalence", -time) %>%
  separate(details, c("population", "strain", "equation"), "\\.") %>%
  mutate(variable = factor(equation, levels=c("z", "w", "y"),
                           labels=c("specific immunity",
                                    "cross-reactive immunity",
                                    "currently infectious")))

ggplot(bind_rows(mutate(timeseries_C, alg="C"),
                 mutate(timeseries_jl, alg="J"))) +
  aes(x=time, y=prevalence, colour=alg) +
  geom_line(alpha=0.5, size=1) +
  facet_grid_sc(rows=vars(variable), cols=vars(strain), scales=list(y=scales_y)) +
  scale_colour_manual(values=my_cols) +
  theme_bw() + theme(legend.position="bottom", legend.title=element_blank())

# ggplot(bind_rows(mutate(timeseries_C, alg="C"),
#                  mutate(timeseries_jl, alg="J"),
#                  mutate(timeseries_R, alg="R")) %>%
#          filter(time > 0.25 * max(time))) +
#   aes(x=time, y=prevalence, colour=alg) +
#   geom_line(alpha=0.5, size=1) +
#   facet_grid_sc(rows=vars(variable), cols=vars(strain), scales=list(y=scales_y)) +
#   # facet_grid_sc(rows=vars(variable), cols=vars(population, strain), scales=list(y=scales_y)) +
#   scale_colour_manual(values=my_cols) +
#   theme_bw() + theme(legend.position="bottom", legend.title=element_blank())
