library(magrittr)
library(tidyverse)
library(JuliaCall)
library(facetscales)
library(patchwork)

julia_setup()
julia_source("../../Code/multipop_MANTIS.jl")

my_cols <- c("#f74700", "#016394", "#b6003b", "#005342")
scales_y <- list(
  `currently infectious (y)`    = scale_y_continuous(limits = c(0, NA)),
  `specific immunity (z)`       = scale_y_continuous(limits = c(0, 1)),
  `cross-reactive immunity (w)` = scale_y_continuous(limits = c(0, 1))
)

## structural parameters
set.seed(0)
struct <- c(2, 2)
strains <- expand.grid(lapply(struct, seq, from=1, by=1) %>%
                         setNames(str_c("locus", 1:length(.))))
n_populations <- 1
# initial_conditions <- rep(0.00001, prod(struct) * n_populations)
initial_conditions <- runif(prod(struct) * n_populations) %>%
  matrix(n_populations, prod(struct)) %>%
  apply(1, . %>% {. / (5 * sum(.))}) %>%
  t()

## dynamical parameters
gamma <- 0.66                            # partial cross-protective immunity (cpi)
sigma <- 8                               # recovery rate
movement_rate <- 0.05
mu <- 0.1
chi <- diag(-movement_rate, 1)
beta <- 2 * (sigma + mu - diag(chi))     # Infection rate

## integration parameters
maxtime <- 1000
tsteps <- round(0.85 * maxtime):maxtime

## run the simulation
timeseries <- julia_call("runMANTIS", strainstructure=struct, tstep=tsteps, tmax=maxtime,
                         beta=beta, gamma=gamma, sigma=sigma, mu=mu, chi=chi,
                         initvals=initial_conditions)$timeseries %>%
  set_colnames(c(expand.grid(str_c("Population_", 1:n_populations),
                             str_c("Strain_", apply(strains, 1, str_c, collapse="")),
                             c("y", "z", "w")) %>% apply(1, str_c, collapse="."),
                 "time")) %>%
  gather("details", "prevalence", -time) %>%
  separate(details, c("population", "strain", "equation"), "\\.") %>%
  mutate(variable = factor(equation, levels=c("y", "z", "w"),
                           labels=c("currently infectious (y)",
                                    "specific immunity (z)",
                                    "cross-reactive immunity (w)")),
         population = factor(population, levels=c("Population_1", "Population_2", "Population_3", "Population_4"),
                             labels=c("Population A", "Population B", "Population C", "Population D"))) %>%
  as_tibble()

p1 <- timeseries %>% filter(equation == "y") %>%
{ggplot(.) +
    aes(colour=strain, y=prevalence, x=time) +
    geom_line(size=1) +
    geom_vline(aes(xintercept=time, linetype=strain), show.legend=FALSE,
               data=group_by(., strain, variable) %>%
                 do(filter(., near(prevalence, max(prevalence), 0.00001))) %>%
                 filter(time - lag(time, default=0) > 20)) +
    facet_grid(strain~variable) +
    scale_x_continuous(expand=c(0,0), breaks=c(860, 900, 940, 980)) +
    scale_colour_manual(values=my_cols) +
    scale_linetype_manual(values=c("Strain_11" = "dashed", "Strain_12" = "dotted", "Strain_21" = "dotted", "Strain_22" = "dashed")) +
    ylab("Proportion of population") +
    theme_bw() +
    theme(legend.position="none",
          strip.text.y=element_blank(),
          strip.background.y=element_blank(),
          axis.title.x=element_blank())}
p2 <- timeseries %>% filter(equation == "z") %>%
{ggplot(.) +
    aes(colour=strain, y=prevalence, x=time) +
    geom_line(size=1) +
    geom_vline(aes(xintercept=time, linetype=strain), show.legend=FALSE,
               data=group_by(., strain, variable) %>%
                 do(filter(., near(prevalence, max(prevalence), 0.0005))) %>%
                 filter(time - lag(time, default=0) > 20)) +
    facet_grid(strain~variable) +
    scale_x_continuous(expand=c(0,0), breaks=c(860, 900, 940, 980)) +
    scale_colour_manual(values=my_cols) +
    scale_y_continuous(limits=0:1) +
    scale_linetype_manual(values=c("Strain_11" = "dashed", "Strain_12" = "dotted", "Strain_21" = "dotted", "Strain_22" = "dashed")) +
    ylab("Proportion of population") +
    theme_bw() +
    theme(legend.position="bottom",
          legend.background=element_blank(),
          legend.title=element_blank(),
          strip.text.y=element_blank(),
          strip.background.y=element_blank(),
          axis.title.y=element_blank())}
p3 <- timeseries %>% filter(equation == "w") %>%
{ggplot(.) +
    aes(colour=strain, y=prevalence, x=time) +
    geom_line(size=0.75) +
    geom_vline(aes(xintercept=time, linetype=strain), show.legend=FALSE,
               data=group_by(., strain, variable) %>%
                 do(filter(., near(prevalence, max(prevalence), 0.001))) %>%
                 filter(time - lag(time, default=0) > 20)) +
    facet_grid(strain~variable) +
    scale_x_continuous(expand=c(0,0), breaks=c(860, 900, 940, 980)) +
    scale_colour_manual(values=my_cols) +
    scale_y_continuous(limits=0:1) +
    scale_linetype_manual(values=c("Strain_11" = "dashed", "Strain_12" = "dotted", "Strain_21" = "dotted", "Strain_22" = "dashed")) +
    ylab("Proportion of population") +
    theme_bw() +
    theme(legend.position="none",
          axis.title.y=element_blank(),
          axis.text.y=element_blank(),
          axis.ticks.y=element_blank(),
          axis.title.x=element_blank())}
p1 + p2 + p3
ggsave(filename="dampening-cycles-all-strains.png", width=9, height=6)
